#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

from __future__ import with_statement

from ionas.fileaccess import (MODE_READ, MODE_WRITE, MODE_APPEND,
                                DEFAULT_LOCALE,)
from ionas.exception import *
from ionas.fileaccess.filelikeobj import (FileLikeObj,)


import os
import sys
import stat
import errno
import locale
import httplib2

from ionas.fileaccess.azureaccess import (azure_except_handler,
                                                 AzureFileObj, AzureAccess)
#from pkgaddon.nifcloud_config import (,)


BUF_SIZE = 65536

RETRY_COUNT = 3


def azure_access_except_handler(path):
    exc_ins = sys.exc_info()[1]
    if isinstance(path, unicode):
        path = path.encode("UTF8")

    azure_except_handler(exc_ins)


def to_open_flag(mode):
    if mode == MODE_READ:
        return "r"
    if mode == MODE_WRITE:
        return "w"
    if mode == MODE_APPEND:
        return "w"
    return "r"


class BackupAzureFileObj(AzureFileObj):
    
    def __renew_file(self):

        count = 0
        while True:
            try:
                if self.file != None:
                    self.file.close()

                self.file = self.accessor.get_renewed_fileobj(
                                self.path, self.mode, self.length)
                return
            except:
                try:
                    azure_access_except_handler(self.path)
                except (RuntimeError, ValueError):
                    if count > RETRY_COUNT:
                        raise
                except (PermissionError, NoSpaceError, NoEntryError):
                    raise
                except:
                    pass

                if count > RETRY_COUNT:
                    raise
                count += 1

    def __call_and_retry_with_renew(self, func, file_like=None):
        count = 0
        while True:
            try:
                return func()
            except:
                try:
                    azure_access_except_handler(self.path)
                except (RuntimeError, ValueError):
                    if count > RETRY_COUNT:
                        raise

                except (PermissionError, NoSpaceError, NoEntryError):
                    raise
                except:
                    pass

                if count > RETRY_COUNT:
                    raise

                self.__renew_file()
                if hasattr(file_like, "seek"):
                    file_like.seek(0)

                count += 1

    def __init__(self, file, path, mode, accessor, length,
            unique_info):

        AzureFileObj.__init__(self, file, path, accessor.access.client,
                                unique_info)
        self.mode = mode
        self.accessor = accessor
        self.length = length

    def __len__(self):

        return int(self.accessor.stat(self.path)[stat.ST_SIZE])

#    def __read(self, size):
#
#        self.point += size
#        return self.file.read(size)
#
#    def __write(self, file_like):
#
#        while True:
#            try:
#                buf = file_like.read(BUF_SIZE)
#            except:
#                azure_access_except_handler(self.path)
#
#            if not buf:
#                break
#            self.file.write(buf)
#
#    def __read_by_path(self, file_like):
#
#        try:
#            file_like.write(self.file)
#        except:
#            azure_access_except_handler(self.path)
#
#    def __write_by_path(self, file_like):
#
#        try:
#            # FIXME 一度に書いて大丈夫か?
#            self.__write(file_like)
#        except:
#            azure_access_except_handler(self.path)
#
    def __close(self):

        try:
            self.file.close()
        except:
            azure_access_except_handler(self.path)

#    def read(self, size):
#
#        func = lambda: self.__read(size)
#        return self.__call_and_retry_with_renew(func)
#
#    def write(self, file_like):
#
#        if self.file == None:
#            open_mode = to_open_flag(self.mode)
#            try:
#                self.length=file_like.get_stat()[stat.ST_SIZE]
#            except:
#                self.length=len(file_like)
#
#            self.__renew_file()
#
#        func = lambda: self.__write(file_like)
#        return self.__call_and_retry_with_renew(func, file_like)
#
#    def write_by_path(self, file_like):
#
#        if self.file == None:
#            open_mode = to_open_flag(self.mode)
#            try:
#                self.length=file_like.get_stat()[stat.ST_SIZE]
#            except:
#                self.length=len(file_like)
#
#            self.__renew_file()
#
#        func = lambda: self.__write_by_path(file_like)
#        return self.__call_and_retry_with_renew(func, file_like)
#
#    def read_by_path(self, file_like):
#
#        func = lambda: self.__read_by_path(file_like)
#        return self.__call_and_retry_with_renew(func, file_like)
#
    def get_stat(self):

        return self.access.stat(self.path)

    def close(self):

        func = lambda: self.__close()
        return self.__call_and_retry_with_renew(func)


class BackupAzureAccess():

    def __call_and_retry(self, func, path):
        count = 0
        while True:
            try:
                return func()
            except:
                try:
                    azure_access_except_handler(path)
                except (RuntimeError, ValueError):
                    if count > RETRY_COUNT:
                        raise

                except (PermissionError, NoSpaceError, NoEntryError):
                    raise
                except:
                    pass

                if count > RETRY_COUNT:
                    raise

                # Retry
                count += 1

    def get_renewed_fileobj(self, path, mode, length):
        """Renew file obj and return it"""

        open_mode = to_open_flag(mode)
        #return self.access.open(path, open_mode, length=length)
        return self.access.open(path, open_mode)

    def _get_server_path(self, path):

        return self.target_info.container + self.sep + path.lstrip(self.sep)

    def __init__(self, target_info, proxy_host=None, proxy_port=None,
            encoding="", unique_info=""):
        self.sep = "/"
        
        if encoding:
            self.encoding = encoding
        else:
            self.encoding = locale.getdefaultlocale()[1]
            if not self.encoding:
                self.encoding = DEFAULT_LOCALE
        self.target_info = target_info

        connection_info = target_info.connection_info
        self.access = AzureAccess(
                        connection_info.accountname,
                        connection_info.accesskey,
                        target_info.container,
                        connection_info.httpmethod,
                        self.encoding,
                        proxy_host=proxy_host, proxy_port=proxy_port,
                        unique_info=unique_info)


    def listdir(self, path):

        marker = None
        contents_list = []

        path = path.strip(self.sep)
        while True:
            _list = self.access.get_list(self.target_info.container,
                                            marker=marker)
            for blob in _list:
                if blob.properties.blob_type != "BlockBlob":
                    continue

                blob_path = blob.name
                prefix = path
                if prefix:
                    prefix += self.sep
                if isinstance(blob_path, unicode):
                    blob_path = blob_path.encode(self.encoding)
                if not prefix or blob_path.startswith(prefix):
                    blob_name = blob_path[len(prefix):].split(self.sep)[0]

                    if blob_name not in contents_list:
                        contents_list.append(blob_name)

            if not _list.next_marker:
                break
            marker = _list.next_marker

        return contents_list

    def isdir(self, entry):  # FIXME 高速化するなら必要かも
        pass

    def open(self, path, mode):
        open_mode = to_open_flag(mode)
        opened_file = None
        if open_mode == "r":
            opened_file = self.access.open(self._get_server_path(path),
                                            open_mode)

        func = lambda: BackupAzureFileObj(opened_file,
                    self._get_server_path(path), mode, accessor=self,
                    length=-1, unique_info=self.access.unique_info)
        return self.__call_and_retry(func, path)

    def stat(self, path):

        func = lambda: self.access.stat(self._get_server_path(path))
        return self.__call_and_retry(func, path)

    def chmod(self, path, mode):
        return NotImplementedError

    def utime(self, path, mtime):
        return NotImplementedError

    def remove(self, path):
        func = lambda: self.access.remove(self._get_server_path(path))
        self.__call_and_retry(func, path)

    def rmdir(self, path):
        func = lambda: self.access.rmdir(self._get_server_path(path))
        self.__call_and_retry(func, path)

    def mkdir(self, path):
        # Do nothing
        pass

    def move(self, src, dst):

        src_parent = src[:dst.rfind(self.sep)]
        dst_parent = dst[:dst.rfind(self.sep)]
        if src_parent != dst_parent:
            func = lambda: self.access.move_directory(src, dst_parent)
            self.__call_and_retry(func, dst)

        new_src= dst_parent + self.sep + src[src.rfind(self.sep)+1:]
        new_name = dst[dst.rfind(self.sep)+1:]
        func = lambda: self.access.rename_directory(new_src, new_name)
        self.__call_and_retry(func, dst)
