# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""Microsoft Azure File Access module.

Define base classes for Microsoft Azure file access.
"""

import stat
import locale
try:
    import email.utils as emailut
except:
    import email.Utils as emailut

from fileaccess import (FileObj, Access)
from ionas.fileaccess import (DEFAULT_LOCALE,)
from ionas.exception import (NoEntryError,)

from azure import (WindowsAzureMissingResourceError,)
from azure.storage import BlobService


class AzureFileObj(FileObj):
    """Class for Azure file."""

    def __init__(self, access, container, path, length=-1):
        """Constructor."""

        self.access = access
        self.container = container
        self.path = path
        self.length = length

        super(AzureFileObj, self).__init__()


    def close(self):
        """Close file."""

        # Default Do nothing
        pass

    def read_to_path(self, file_path):
        """Read data and write to a "file_path" file."""

        if isinstance(file_path, unicode):
            file_path = file_path.encode("UTF8")
        self.access.client.get_blob_to_path(self.container, self.path,
                                                file_path)

    def write_from_path(self, file_path):
        """Read data from a "file_path" file and write it."""

        if isinstance(file_path, unicode):
            file_path = file_path.encode("UTF8")
        self.access.client.put_block_blob_from_path(self.container, self.path,
                                                        file_path)


class AzureAccess(Access):
    """Class accessing to Azure file."""

    def _to_server_path(self, path, is_dir=False):

        if isinstance(path, unicode):
            path = path.encode(self.encoding)

        path = path.strip(self.sep)
        if is_dir and path:
            return path + self.sep
        return path


    def _get_time(self, blob):

        return emailut.mktime_tz(emailut.parsedate_tz(blob["last-modified"]))

    def __init__(self, account_name, account_key, container="",
        protocol="https", encoding="", proxy_host=None, proxy_port=None,
        unique_info=""):
        """Constructor."""

        self.sep = "/"
        if encoding:
            self.encoding = encoding
        else:
            self.encoding = locale.getdefaultlocale()[1]
            if not self.encoding:
                self.encoding = DEFAULT_LOCALE

        self.container = container
        self.unique_info = unique_info

        self.client = BlobService(account_name, account_key, protocol)
        if proxy_host:
            self.client.set_proxy(proxy_host, proxy_port)

    def make_directory(self, dir_path):
        """Create "dir_path" directory."""

        # Do nothing
        pass

    def delete_directory(self, dir_path):
        """Delete "dir_path" directory."""

        server_path = self._to_server_path(dir_path, True)
        marker = None
        while True:
            _list = self.client.list_blobs(self.container,
                                prefix=server_path, marker=marker)
            for blob in _list:
                if blob.properties.blob_type != "BlockBlob":
                    continue
                if blob.name[-1] == self.sep:
                    continue
                self.client.delete_blob(self.container, blob.name)
            if not _list.next_marker:
                break
            marker = _list.next_marker

    def stat_directory(self, dir_path):
        """Return a stat information dictionary for "dir_path"
           which keys below:
            stat.ST_MODE, stat.ST_MTIME, stat.ST_SIZE
        """

        try:
            blob = self.client.get_blob_properties(self.container,
                                    self._to_server_path(dir_path, True))
            return {
                    stat.ST_MODE: stat.S_IFDIR,
                    stat.ST_MTIME: self._get_time(blob),
                    stat.ST_SIZE: int(blob["content-length"]),
                }
        except:
            return {
                    stat.ST_MODE: stat.S_IFDIR,
                    stat.ST_MTIME: 0,
                    stat.ST_SIZE: 0,
                }
            
    def get_contents(self, dir_path, with_stat=False):
        """Return a contents dictionary in "dir_path".
           A stat information must be included, if "with_stat" is True, 
        """

        marker = None
        contents_dict = {}

        server_path = self._to_server_path(dir_path, True)
        prefix = server_path
        prefix_len = len(dir_path)
        if not prefix:
            prefix_len = 0
            prefix = None

        while True:
            _list = self.client.list_blobs(self.container,
                                prefix=prefix, marker=marker)
            for blob in _list:
                if blob.properties.blob_type != "BlockBlob":
                    continue

                blob_name = blob.name
                if isinstance(blob_name, str):
                    blob_name = blob_name.decode(self.encoding)

                child_name = blob_name[prefix_len:].strip(self.sep).split(
                                                                self.sep)[0]
                if child_name in contents_dict:
                    continue

                if not with_stat:
                    contents_dict[child_name] = {}
                    continue

                userver_path = server_path
                if isinstance(userver_path, str):
                    userver_path = userver_path.decode(self.encoding)
                child_path = userver_path + child_name
                if blob_name.startswith(child_path + self.sep):
                    contents_dict[child_name] = self.stat_directory(child_path)
                elif blob_name == child_path:
                    contents_dict[child_name] = self.stat_file(child_path)

            if not _list.next_marker:
                break
            marker = _list.next_marker

        return contents_dict

    def open(self, file_path, mode="r", length=-1):
        """Return a class instance of the "FileObj" or sub class
           corresponding to "file_path".
        """

        return AzureFileObj(self, self.container,
                self._to_server_path(file_path), length)

    def delete_file(self, file_path):
        """Delete "file_path" file."""

        self.client.delete_blob(self.container, self._to_server_path(file_path))

    def stat_file(self, file_path):
        """Return a stat information dictionary for "file_path"
           which keys below:
            stat.ST_MODE, stat.ST_MTIME, stat.CTIME, stat.ST_SIZE
        """

        blob = self.client.get_blob_properties(self.container,
                                self._to_server_path(file_path))
        return {
                stat.ST_MODE: 0,
                stat.ST_MTIME: self._get_time(blob),
                stat.ST_SIZE: int(blob["content-length"]),
            }
